@@ -26,11 +26,13 @@
 (require 'pdf-info)
 (require 'pdf-cache)
 (require 'pdf-misc)
+(require 'pdf-util)
 (require 'facemenu) ;; list-colors-duplicates
 (require 'faces) ;; color-values
 (require 'org)   ;; org-create-formula-image
 (require 'tablist)
 (require 'cl-lib)
+(require 'seq)
 
 
 ;; * ================================================================== *
@@ -477,6 +479,7 @@ Signals an error, if PROPERTY is not modifiable.
 Returns the modified annotation."
 
   (declare (indent 2))
+  (setq a (pdf-annot--ensure-fresh a))
   (unless (equal value (pdf-annot-get a property))
     (unless (pdf-annot-property-modifiable-p a property)
       (error "Property `%s' is read-only for this annotation"
@@ -556,6 +559,15 @@ have identical id properties."
   "Return id property of annotation A."
   (pdf-annot-get a 'id))
 
+(defun pdf-annot--ensure-fresh (a)
+  "Return a fresh version of annotation A from the server.
+If the annotation does not exist anymore, signal an error."
+  (let ((id (pdf-annot-get-id a)))
+    (or (cl-find id
+                 (pdf-annot-getannots (pdf-annot-get a 'page) nil (pdf-annot-get-buffer a))
+                 :key #'pdf-annot-get-id)
+        (user-error "No such annotation: %s" id))))
+
 (defun pdf-annot-get-type (a)
   "Return type property of annotation A."
   (pdf-annot-get a 'type))
@@ -579,6 +591,7 @@ This function always returns nil."
   (interactive
    (list (pdf-annot-read-annotation
           "Click on the annotation you wish to delete")))
+  (setq a (pdf-annot--ensure-fresh a))
   (with-current-buffer (pdf-annot-get-buffer a)
     (pdf-info-delannot
      (pdf-annot-get-id a))
@@ -642,6 +655,7 @@ The DO-SAVE argument is given to
 `pdf-info-getattachment-from-annot', which see."
   (unless (pdf-annot-has-attachment-p a)
     (error "Annotation has no data attached: %s" a))
+  (setq a (pdf-annot--ensure-fresh a))
   (pdf-info-getattachment-from-annot
    (pdf-annot-get-id a)
    do-save
@@ -890,6 +904,7 @@ i.e. a non mouse-movement event is read."
   (let* ((mpos (posn-object-x-y (event-start event)))
          (a (or annot
                 (pdf-annot-at-position mpos))))
+    (setq a (pdf-annot--ensure-fresh a))
     (unless a
       (error "No annotation at this position: %s" mpos))
     (let* ((apos (pdf-annot-image-position a))
@@ -997,7 +1012,8 @@ other annotations."
                 `("white" "steel blue" 0.35 ,@edges))
              :map (pdf-view-apply-hotspot-functions
                    window page size)
-             :width (car size))))
+             :width (car size))
+           (when pdf-view-roll-minor-mode page)))
         (pdf-util-scroll-to-edges
          (pdf-util-scale-relative-to-pixel (car edges)))))))
 
@@ -1053,9 +1069,12 @@ Return the new annotation."
              (> (length edges) 1))
     (error "Edges argument should be a single edge-list for text annotations"))
   (let* ((selection-style pdf-view-selection-style)
+         (non-markup (pcase type
+                       ('text t)
+                       ('highlight pdf-view--have-rectangle-region)))
          (a (apply #'pdf-info-addannot
                    page
-                   (if (eq type 'text)
+                   (if non-markup
                        (car edges)
                      (apply #'pdf-util-edges-union
                             (apply #'append
@@ -1066,7 +1085,7 @@ Return the new annotation."
                    type
                    selection-style
                    nil
-                   (if (not (eq type 'text)) edges)))
+                   (unless non-markup edges)))
          (id (pdf-annot-get-id a)))
     (when property-alist
       (condition-case err
@@ -1081,8 +1100,8 @@ Return the new annotation."
       (pdf-annot-activate-annotation a))
     a))
 
-(defun pdf-annot-add-text-annotation (pos &optional icon property-alist)
-  "Add a new text annotation at POS in the selected window.
+(defun pdf-annot-add-text-annotation (pos &optional icon property-alist page)
+  "Add a new text annotation at POS on PAGE in the selected window.
 
 POS should be a image position object or a cons \(X . Y\), both
 being image coordinates.
@@ -1095,7 +1114,7 @@ Adjust X and Y accordingly, if the position would render the
 annotation off-page.
 
 Merge ICON as a icon property with PROPERTY-ALIST and
-`pdf-annot-default-text-annotation-properties' and apply the
+`pdf-annot-default-annotation-properties' and apply the
 result to the created annotation.
 
 See also `pdf-annot-add-annotation'.
@@ -1110,6 +1129,9 @@ Return the new annotation."
      (list posn)))
   (pdf-util-assert-pdf-window)
   (when (posnp pos)
+    (setq page (or page
+                   (when pdf-view-roll-minor-mode
+                     (1+ (/ (posn-point pos) 4)))))
     (setq pos (posn-object-x-y pos)))
   (let ((isize (pdf-view-image-size))
         (x (car pos))
@@ -1132,10 +1154,10 @@ Return the new annotation."
        (pdf-annot-merge-alists
         (and icon `((icon . ,icon)))
         property-alist
-        pdf-annot-default-text-annotation-properties
         (cdr (assq 'text pdf-annot-default-annotation-properties))
         (cdr (assq t pdf-annot-default-annotation-properties))
-        `((color . ,(car pdf-annot-color-history))))))))
+        `((color . ,(car pdf-annot-color-history))))
+       page))))
 
 (defun pdf-annot-mouse-add-text-annotation (ev)
   "Add a text annotation using the mouse.
@@ -1151,18 +1173,19 @@ EV describes the captured mouse event."
           "Click where a new text annotation should be added ..."))
      (event-start ev))))
 
-(defun pdf-annot-add-markup-annotation (list-of-edges type &optional color
+(defun pdf-annot-add-markup-annotation (region type &optional color
                                                       property-alist)
   "Add a new markup annotation in the selected window.
 
-LIST-OF-EDGES determines the marked up area and should be a list
+REGION determines the marked up area and should be a cons cell
+\(PAGE . LIST-OF-EDGES\) where LIST-OF-EDGES should be list
 of \(LEFT TOP RIGHT BOT\), each value a relative coordinate.
 
 TYPE should be one of `squiggly', `underline', `strike-out' or
 `highlight'.
 
 Merge COLOR as a color property with PROPERTY-ALIST and
-`pdf-annot-default-markup-annotation-properties' and apply the
+`pdf-annot-default-annotation-properties' and apply the
 result to the created annotation.
 
 See also `pdf-annot-add-annotation'.
@@ -1178,17 +1201,16 @@ Return the new annotation."
   (pdf-util-assert-pdf-window)
   (pdf-annot-add-annotation
    type
-   list-of-edges
+   (cdr region)
    (pdf-annot-merge-alists
     (and color `((color . ,color)))
     property-alist
-    pdf-annot-default-markup-annotation-properties
     (cdr (assq type pdf-annot-default-annotation-properties))
     (cdr (assq t pdf-annot-default-annotation-properties))
     (when pdf-annot-color-history
       `((color . ,(car pdf-annot-color-history))))
     '((color . "#ffff00")))
-   (pdf-view-current-page)))
+   (car region)))
 
 (defun pdf-annot-add-squiggly-markup-annotation (list-of-edges
                                                  &optional color property-alist)
@@ -1235,12 +1257,10 @@ properties. See also `pdf-annot-add-markup-annotation'."
 
 Offer `pdf-annot-color-history' as default values."
   (let* ((defaults (append
-                    (delq nil
-                          (list
-                           (cdr (assq 'color
-                                      pdf-annot-default-markup-annotation-properties))
-                           (cdr (assq 'color
-                                      pdf-annot-default-text-annotation-properties))))
+                    (seq-map #'cdr
+                             (seq-filter (lambda (x) (eq 'color (car x)))
+                                         (seq-mapcat #'cdr
+                                                     pdf-annot-default-annotation-properties)))
                     pdf-annot-color-history))
          (prompt
           (format "%s%s: "
@@ -1447,15 +1467,16 @@ annotation's contents and otherwise `org-mode'."
     (set-keymap-parent kmap text-mode-map)
     (define-key kmap (kbd "C-c C-c") #'pdf-annot-edit-contents-commit)
     (define-key kmap (kbd "C-c C-q") #'pdf-annot-edit-contents-abort)
+    (define-key kmap (kbd "C-c C-k") #'pdf-annot-edit-contents-abort)
     kmap))
 
 (define-minor-mode pdf-annot-edit-contents-minor-mode
   "Active when editing the contents of annotations."
   :group 'pdf-annot
   (when pdf-annot-edit-contents-minor-mode
-    (message "%s"
-             (substitute-command-keys
-              "Press \\[pdf-annot-edit-contents-commit] to commit your changes, \\[pdf-annot-edit-contents-abort] to abandon them."))))
+    (setq-local header-line-format
+                (substitute-command-keys "\
+Press \\[pdf-annot-edit-contents-commit] to commit your changes, \\[pdf-annot-edit-contents-abort] to abandon them."))))
 
 (put 'pdf-annot-edit-contents-minor-mode 'permanent-local t)
 
@@ -1508,13 +1529,11 @@ At any given point of time, only one annotation can be in edit mode."
                (not (eq a pdf-annot-edit-contents--annotation)))
       (with-current-buffer pdf-annot-edit-contents--buffer
         (pdf-annot-edit-contents-finalize 'ask)))
+    (setq a (pdf-annot--ensure-fresh a))
     (unless (buffer-live-p pdf-annot-edit-contents--buffer)
       (setq pdf-annot-edit-contents--buffer
-            (with-current-buffer (get-buffer-create
-                                  (format "*Edit Annotation %s*"
-                                          (buffer-name)))
-              (pdf-annot-edit-contents-minor-mode 1)
-              (current-buffer))))
+            (get-buffer-create
+             (format "*Edit Annotation %s*" (buffer-name)))))
     (with-current-buffer pdf-annot-edit-contents--buffer
       (let ((inhibit-read-only t))
         (erase-buffer)
@@ -1522,6 +1541,7 @@ At any given point of time, only one annotation can be in edit mode."
         (set-buffer-modified-p nil))
       (setq pdf-annot-edit-contents--annotation a)
       (funcall pdf-annot-edit-contents-setup-function a)
+      (pdf-annot-edit-contents-minor-mode 1)
       (current-buffer))))
 
 (defun pdf-annot-edit-contents (a)
@@ -1540,6 +1560,66 @@ At any given point of time, only one annotation can be in edit mode."
       (error "No annotation at this position"))
     (pdf-annot-edit-contents a)))
 
+(defun pdf-annot-edit (annot)
+  "Activate ANNOT, for editing.
+
+Interactively, annot is read via `pdf-annot-read-annot'.
+This function displays characters around the annots in the current
+page and starts reading characters (ignoring case).  After a
+sufficient number of characters have been read, the corresponding
+annot's annot is invoked.  Additionally, SPC may be used to
+scroll the current page."
+  (interactive
+   (list (or (pdf-annot-read-annot "Activate annot (SPC scrolls): ")
+             (error "No annot selected"))))
+  (pdf-annot-activate-annotation annot))
+
+;; TODO 'merge' this function with `pdf-links-read-link-action' into a single
+;; universal 'read-action' function (in `pdf-util'?)
+(defun pdf-annot-read-annot (prompt)
+  "Using PROMPT, interactively read an annot-action.
+
+See `pdf-annot-edit' for the interface."
+  (pdf-util-assert-pdf-window)
+  (let* ((annots (pdf-annot-getannots (pdf-view-current-page) nil nil))
+         (keys (pdf-links-read-link-action--create-keys
+                (length annots)))
+         (key-strings (mapcar (apply-partially 'apply 'string)
+                              keys))
+         (alist (cl-mapcar 'cons keys annots))
+         (size (pdf-view-image-size))
+         (colors (pdf-util-face-colors
+                  'pdf-links-read-link pdf-view-dark-minor-mode))
+         (args (list
+                :foreground (car colors)
+                :background (cdr colors)
+                :formats
+                `((?c . ,(lambda (_edges) (pop key-strings)))
+                  (?P . ,(number-to-string
+                          (max 1 (* (cdr size)
+                                    pdf-links-convert-pointsize-scale)))))
+                :commands pdf-links-read-link-convert-commands
+                :apply (pdf-util-scale-relative-to-pixel
+                        (mapcar (lambda (l) (cdr (assq 'edges l)))
+                                annots)))))
+    ;; (print (plist-get args :apply))
+    (unless annots
+      (error "No annots on this page"))
+    (unwind-protect
+        (let ((image-data
+               (pdf-cache-get-image
+                (pdf-view-current-page)
+                (car size) (car size) 'pdf-annot-read-annot)))
+          (unless image-data
+            (setq image-data (apply 'pdf-util-convert-page args ))
+            (pdf-cache-put-image
+             (pdf-view-current-page)
+             (car size) image-data 'pdf-annot-read-annot))
+          (pdf-view-display-image
+           (create-image image-data (pdf-view-image-type) t)
+           (when pdf-view-roll-minor-mode (pdf-view-current-page)))
+          (pdf-links-read-link-action--read-chars prompt alist))
+      (pdf-view-redisplay))))
 
 
 ;; * ================================================================== *
@@ -1582,6 +1662,7 @@ Currently supported properties are page, type, label, date and contents."
 
 (defvar pdf-annot-list-mode-map
   (let ((km (make-sparse-keymap)))
+    (define-key km (kbd "e") 'pdf-annot-edit)
     (define-key km (kbd "C-c C-f") #'pdf-annot-list-follow-minor-mode)
     (define-key km (kbd "SPC") #'pdf-annot-list-display-annotation-from-id)
     km))
@@ -1720,16 +1801,19 @@ pretty-printed output."
 \\{pdf-annot-list-mode-map}"
   (interactive)
   (pdf-util-assert-pdf-buffer)
-  (let ((buffer (current-buffer)))
-    (with-current-buffer (get-buffer-create
-                          (format "*%s's annots*"
-                                  (file-name-sans-extension
-                                   (buffer-name))))
+  (let* ((buffer (current-buffer))
+         (name (format "*%s's annots*"
+                       (file-name-sans-extension
+                        (buffer-name))))
+         (annots-existed (and (get-buffer name)
+                              pdf-annot-list-buffer)))
+    (with-current-buffer (get-buffer-create name)
       (delay-mode-hooks
         (unless (derived-mode-p 'pdf-annot-list-mode)
           (pdf-annot-list-mode))
         (setq pdf-annot-list-document-buffer buffer)
-        (tabulated-list-print)
+        (unless annots-existed
+          (tabulated-list-print))
         (setq tablist-context-window-function
               (lambda (id) (pdf-annot-list-context-function id buffer))
               tablist-operations-function #'pdf-annot-list-operation-function)
