@@ -51,6 +51,13 @@
 (require 'tq)
 (require 'cl-lib)
 
+;; Silence native-comp warnings about functions defined elsewhere
+(declare-function pdf-util-frame-scale-factor "pdf-util")
+(declare-function pdf-util-hexcolor "pdf-util")
+(declare-function pdf-util-munch-file "pdf-util")
+(declare-function pdf-util-highlight-regexp-in-string "pdf-util")
+(declare-function pdf-view-buffer-file-name "pdf-view")
+
 
 
 ;; * ================================================================== *
@@ -572,6 +579,8 @@ interrupted."
            (cl-case key
              ((:render/printed)
               (setq value (equal value "1")))
+             ((:render/gammabeforeinvert)
+              (setq value (equal value "1")))
              ((:render/usecolors)
               (setq value (ignore-errors
                             (let ((int-val (cl-parse-integer value)))
@@ -1734,6 +1743,8 @@ Returns a list \(LEFT TOP RIGHT BOT\)."
                    soptions))
             ((:render/printed)
              (push (if value 1 0) soptions))
+            ((:render/gammabeforeinvert)
+             (push (if value 1 0) soptions))
             ((:render/usecolors)
              ;; 0 -> original color
              ;; 1 -> recolor document to grayscale mapping black to
