@@ -143,6 +143,19 @@
     ["Copy region" pdf-view-kill-ring-save
      :keys "\\[kill-ring-save]"
      :active (pdf-view-active-region-p)]
+    ("Selection style"
+     ["Glyph" (pdf-view-set-selection-style 'glyph)
+      :style radio
+      :selected (eq pdf-view-selection-style 'glyph)
+      :help "When dragging the mouse, select individual characters."]
+     ["Word" (pdf-view-set-selection-style 'word)
+      :style radio
+      :selected (eq pdf-view-selection-style 'word)
+      :help "When dragging the mouse, select entire words."]
+     ["Line" (pdf-view-set-selection-style 'line)
+      :style radio
+      :selected (eq pdf-view-selection-style 'line)
+      :help "When dragging the mouse, select entire lines."])
     "--"
     ["Isearch document" isearch-forward
      :visible (bound-and-true-p pdf-isearch-minor-mode)]
@@ -199,6 +212,9 @@
                    (lookup-key pdf-misc-menu-bar-minor-mode-map
                                [menu-bar pdf\ tools]))))))
 
+(define-derived-mode pdf-metadata-buffer-mode special-mode "PDF Metadata"
+  "View and traverse the metadata of a PDF file.")
+
 (defun pdf-misc-display-metadata ()
   "Display all available metadata in a separate buffer."
   (interactive)
@@ -206,6 +222,8 @@
   (let* ((buffer (current-buffer))
          (md (pdf-info-metadata)))
     (with-current-buffer (get-buffer-create "*PDF-Metadata*")
+      (unless (derived-mode-p 'pdf-metadata-buffer-mode)
+        (pdf-metadata-buffer-mode))
       (let* ((inhibit-read-only t)
              (pad (apply' max (mapcar (lambda (d)
                                         (length (symbol-name (car d))))
